<?php
namespace Service\Manager\Mongo;

use Service\Manager\Mongo\Query\Expr;
use Service\Manager\Mongo\Types\Type;
use Util\Utilities;
use Zeedhi\Framework\DataSource\AssociatedWithDataSource;
use Zeedhi\Framework\DataSource\Configuration;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DataSource\Manager\Doctrine\NameProvider;
use Zeedhi\Framework\DB\Mongo\Mongo;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;
use Zeedhi\Zmart\QueryBuilders\Mongo\QueryBuilder;
use Zeedhi\Framework\DB\Mongo\Exception;
use Zeedhi\Framework\DependencyInjection\InstanceManager;
use MongoDB\BSON\Regex;

class Manager implements \Zeedhi\Framework\DataSource\Manager {

    const PRIMARY_KEY_COLUMN = 'id';
    const MONGO_KEY_COLUMN = '_id';
    /** @var array */
    protected $propertiesToClear = array(
        '__createdLocal',
        '__is_new',
        '__isSelected',
        'changed',
        '__index',
        '__notSelectable'
    );
    /** @var \Zmart\Mongo\Mongo */
    public $mongo;
    /** @var NameProvider */
    protected $nameProvider;
    /** @var Configuration */
    protected $dataSourceConfig;
    /** @var QueryBuilder */
    protected $queryBuilder;

    protected $dbName;

    /**
     * @param Mongo $mongo
     * @param NameProvider $nameProvider
     * @param QueryBuilder $queryBuilder
     */
    public function __construct(Mongo $mongo, NameProvider $nameProvider, QueryBuilder $queryBuilder) {
        $this->mongo = $mongo;
        $this->nameProvider = $nameProvider;
        $this->queryBuilder = $queryBuilder;
    }


    /**
     * Persist all given rows in DataSet.
     *
     * @param DataSet $dataSet The collection and description of rows.
     *
     * @return array Rows with primary key columns values.
     *
     * @throws
     */
    public function persist(DataSet $dataSet) {
        $this->loadCurrentDataSource($dataSet);
        return $this->persistRows($dataSet);
    }
    public function insertMany(string $collectionName, array $objects) {
        return $this->mongo->insertMany($collectionName,$objects);
    }

    /**
     * @param FilterCriteria $filterCriteria
     *
     * @return DataSet
     */
    public function findBy(FilterCriteria $filterCriteria) {
        $this->loadCurrentDataSource($filterCriteria);
        $rows = $this->retrieveRows($filterCriteria);
        return new DataSet($filterCriteria->getDataSourceName(), $rows);
    }

    /**
     * Perform an aggregate query
     * @param ZmartFilterCriteria $filterCriteria
     * @return DataSet
     */
    public function aggregate(ZmartFilterCriteria $filterCriteria, $allowDivideByZero = false) {
        $this->loadCurrentDataSource($filterCriteria);
        $rows = $this->retrieveRowsAggregated($filterCriteria, $allowDivideByZero);
        return new DataSet($filterCriteria->getDataSourceName(), $rows);
    }

    /**
     * Delete all given rows in DataSet.
     *
     * @param DataSet $dataSet The collection and description of rows.
     *
     * @return array Rows with primary key columns values.
     *
     */
    public function delete(DataSet $dataSet) {
        $this->loadCurrentDataSource($dataSet);
        return $this->deleteRows($dataSet);
    }

    public function remove(string $collectionName, FilterCriteria  $filterCriteria, bool $justOne = false) {
        $criteria = $this->processNormalFilterConditions($filterCriteria);
        return $this->mongo->remove($collectionName, $criteria, $justOne);
    }

    protected function deleteRow($row) {
        $criteria = $this->buildCriteriaFromRow($row);
        $this->mongo->remove($this->dataSourceConfig->getTableName(), $criteria);
    }

    public function deleteMany($collectionName, $criteria) {
        $this->mongo->remove($collectionName, $criteria, false);
    }

    /**
     * @param DataSet $dataSet
     *
     * @return array
     */
    protected function deleteRows(DataSet $dataSet) {
        $deletedRows = array();
        foreach ($dataSet->getRows() as $key => $row) {
            $this->deleteRow($row);
            $deletedRows[$key] = $this->getPrimaryKeyValueFromRow($row[self::MONGO_KEY_COLUMN]);
        }
        return $deletedRows;
    }

    /**
     * @param AssociatedWithDataSource $associatedWithDataSource
     */
    protected function loadCurrentDataSource(AssociatedWithDataSource $associatedWithDataSource) {
        $this->dataSourceConfig = $this->nameProvider->getDataSourceByName($associatedWithDataSource->getDataSourceName());
    }

    /**
     * @param $row
     * @return array
     */
    protected function buildCriteriaFromRow($row) {
        $expression = new Expr();
        $expression->field(self::MONGO_KEY_COLUMN)->equals(Type::getType('id')->convertToDatabaseValue($row[self::MONGO_KEY_COLUMN]));
        return $expression->getQuery();
    }

    /**
     * @param $condition
     * @return array
     */
    protected function parseToDatabaseCondition($condition) {
        if ($condition['columnName'] === self::MONGO_KEY_COLUMN) {
            if (is_array($condition['value'])) {
                $ids = array();
                foreach ($condition['value'] as $value)
                    $ids[] = new \MongoDB\BSON\ObjectID($value);
                return $ids;
            }
            return new \MongoDB\BSON\ObjectID($condition['value']);
        }
        return $condition['value'];
    }

    /**
     * @param $row
     * @return array
     */
    protected function getPrimaryKeyValueFromRow($mongoId) {
        return array(
            self::MONGO_KEY_COLUMN => (string)$mongoId
        );
    }

    /**
     * @param DataSet $dataSet
     * @return array
     */
    protected function persistRows(DataSet $dataSet) {
        $persistedRows = array();
        foreach ($dataSet->getRows() as $key => $row) {
            $mongoId = $this->persistRow($row);
            $persistedRows[$key] = $this->getPrimaryKeyValueFromRow($mongoId);
        }

        return $persistedRows;
    }

    /**
     * @param $row
     */
    protected function persistRow($row) {
        $row = $this->clearRowBeforePersist($row);
        $mongoId = null;
        if (isset($row[self::MONGO_KEY_COLUMN])) {
            $documentKey = $row[self::MONGO_KEY_COLUMN];
            $criteria = $this->buildCriteriaFromRow($row);
            unset($row[self::MONGO_KEY_COLUMN]);
            $updateMongo = array('$set' => $row);
            $this->mongo->update($this->dataSourceConfig->getTableName(), $criteria, $updateMongo);
            $row[self::MONGO_KEY_COLUMN] = $documentKey;
            $mongoId = $documentKey;
        } else {
            $insertMongo = $this->mongo->insert($this->dataSourceConfig->getTableName(), $row);
            $mongoId = $insertMongo->getInsertedId();
        }
        return $mongoId;
    }

    /**
     * @param FilterCriteria $filterCriteria
     * @param null $columns
     * @return array
     */
    protected function processNormalFilterConditions(FilterCriteria $filterCriteria, $columns = null) {
        $conditions = $filterCriteria->getConditions();
        foreach($conditions as &$condition) {
            $condition['value'] = $this->parseToDatabaseCondition($condition);
        }
        $whereClause = $this->queryBuilder->getWhere($conditions, $columns);
        return $whereClause ?: array();
    }

    /**
     * @param ZmartFilterCriteria $filterCriteria
     * @return array
     */
    protected function processAggregateFilterConditions(ZmartFilterCriteria $filterCriteria, $allowDivideByZero = false) {
        return $this->queryBuilder->buildQueryFromFilterCriteria($filterCriteria, $allowDivideByZero);
    }

    /**
     * @param FilterCriteria $filterCriteria
     * @return array
     */
    public function processSort(FilterCriteria $filterCriteria) {
        $sort = $this->queryBuilder->getSort($filterCriteria->getOrderBy());
        return $sort ?: array();
    }

    /**
     * @param $row
     * @return mixed
     */
    protected function planningDocumentKey($row) {
        if(isset($row[self::MONGO_KEY_COLUMN])) {
            $row[self::MONGO_KEY_COLUMN] = Type::convertDatabaseValueToPHP($row[self::MONGO_KEY_COLUMN]);
        }
        foreach($row as &$rowProperty) {
            if(is_array($rowProperty)) {
                $rowProperty = $this->planningDocumentKey($rowProperty);
            } else if($rowProperty instanceof \MongoDB\BSON\UTCDateTime) {
                $rowProperty = Type::convertDatabaseValueToPHP($rowProperty);
            }
        }
        return $row;
    }

    /**
     * @param $row
     * @return mixed
     */
    protected function clearRowBeforePersist($row) {
        foreach ($this->propertiesToClear as $property) {
            if (isset($row[$property])) {
                unset($row[$property]);
            }
        }
        return $row;
    }

    /**
     * @param $row
     * @return array
     */
    protected function filterColumnsForResultSet($row) {
        $columnsForResultSet = array();
        if($this->dataSourceConfig){
            $columnsForResultSet = $this->dataSourceConfig->getColumnsForResultSet();
        }
        if (count($columnsForResultSet) > 0) {
            $rowFiltered = array();
            foreach ($columnsForResultSet as $columnName) {
                if (array_key_exists($columnName, $row)) {
                    $rowFiltered[$columnName] = $row[$columnName];
                }
            }
            return $rowFiltered;
        }
        return $row;
    }

    /**
     * @param $rows
     * @return array
     */
    protected function prepareRowsForResultSet($rows) {
        $preparedRows = array();
        foreach ($rows as $row) {
            $row = $this->planningDocumentKey($row);
            $preparedRows[] = $this->filterColumnsForResultSet($row);
        }
        return $preparedRows;
    }

    public function simpleFindPagination(FilterCriteria $filterCriteria){
        $criteria = $this->processNormalFilterConditions($filterCriteria);
        $limit = $filterCriteria->getPageSize();
        $page = $filterCriteria->getPage();
        $numSkip = $limit * ($page - 1);
        $sort = $this->processSort($filterCriteria);
        return $this->mongo->findPagination($filterCriteria->getDataSourceName(), $criteria, $sort , $limit, $numSkip);
    }

    public function count(FilterCriteria $filterCriteria){
        $criteria = $this->processNormalFilterConditions($filterCriteria);
        return $this->mongo->count($filterCriteria->getDataSourceName(), $criteria);
    }

    /**
     * @param FilterCriteria $filterCriteria
     * @return array
     */
    protected function retrieveRows(FilterCriteria $filterCriteria) {
        $criteria = $this->processNormalFilterConditions($filterCriteria);
        $sort = $this->processSort($filterCriteria);
        $limit = $filterCriteria->getPageSize();
        $rows = $this->mongo->find($filterCriteria->getDataSourceName(), $criteria, $sort , $limit);
        return $this->prepareRowsForResultSet($rows);
    }

    /**
     * @param ZmartFilterCriteria $filterCriteria
     * @return array
     * @throws \Zeedhi\Framework\DB\Mongo\Exception
     */
    protected function retrieveRowsAggregated(ZmartFilterCriteria $filterCriteria, $allowDivideByZero = false) {
        $criteria = $this->processAggregateFilterConditions($filterCriteria, $allowDivideByZero);
        $rows = $this->managerAggregate($filterCriteria->getDataSourceName(), $criteria);
        return $this->prepareRowsForResultSet($rows);
    }

     public function managerAggregate($collectionName, array $criteria) {
        $collection = $this->mongo->getCollection($collectionName);
        try {
            $cursor = $collection->aggregate($criteria,array('timeout'=>90000));
           // $cursor->timeout(90000);
            return iterator_to_array($cursor);
        } catch (\MongoException $e) {
            throw Exception::aggregateError($e);
        }
     }

    /**
     * Get name of all collections in a database.
     *
     * @return array
     *
     */
    public function getCollectionNames(){
        return $this->mongo->getCollectionNames();
    }

    public function getCollection($collectionName){
        return $this->mongo->getCollection($collectionName);
    }

    /**
     * Drop collection by name
     * @param $collectionName
     */
    public function dropCollection($collectionName){
        $this->mongo->dropCollection($collectionName);
    }

    /**
     * Create a database by name
     * @param $collectionName
     */
    public function createDatabase($dbName){
        $this->mongo->setDbByName($dbName);
        $this->dbName = $dbName;
    }

    /**
     * Get current dbname
     */
    public function getDbName(){
        return $this->dbName;
    }

    /**
     * Create a user db
     * @param $db_Name
     */
    public function createDbUser($db_name)
    {
        $command = array(
            "createUser" => "zmartzeedhi",
            "pwd" => "zmartzeedhi",
            "roles" => array(
                array(
                    "role" => "readWrite",
                    "db" => $db_name
                )
            )
        );
        $this->mongo->command($command);
    }
    public function renameCollection($odd,$new){
        $newManager = $this->buildManager('admin');
        $command=array(
            "renameCollection" => 'Zmart.' . $odd,
            "to"=> 'Zmart.' . $new
        );
        $newManager->command($command);
    }

    public function cloneCollectionToImportDatabase($oldId, $newId){
        try{
            $command = array("drop" => $newId);
            try{
                $this->mongo->command($command);
            }catch(\Exception $e){

            }
            $oldDbname = Utilities::getEnvironment()->getBaseConnection()[0]["dbName"];
            $newDbname = "Zmart__5";
            $cloneSuffix = "_clone";
            $newManager = $this->buildManager($oldDbname);
            $newManager->mongo->getCollection($oldId)->aggregate(
                array(array('$out'=> $oldId.$cloneSuffix) )
            );
            $newManager = $this->buildManager('admin');
            $command=array(
                "renameCollection" => $oldDbname.".".$oldId.$cloneSuffix,
                "to"=> $newDbname.".".$newId
            );
            $newManager->command($command);
        }catch(\Exception $e){
            var_dump($e);
        }
    }

    public function buildManager($dbName){
        $server = InstanceManager::getInstance()->getParameter('mongoHost');
        $port =  InstanceManager::getInstance()->getParameter('mongoPort');
        $mongoOptions = array("socketTimeoutMS" => 1800000);

        $im = \Zeedhi\Framework\DependencyInjection\InstanceManager::getInstance();

        $mongo = new Mongo($server, $port, $dbName, $mongoOptions);
        $newManager = new Manager($mongo, $im->getService("nameProvider"), $im->getService("mongoQueryBuilder"));

        return $newManager;
    }

    public function cloneDBName($dbNameOld,$dbName,$mongoHost){
        $command =
            array(
                    'copydb' => 1,
                    'fromdb' => $dbNameOld,
                    'todb' => $dbName
                );

        if($mongoHost){
            $command['fromhost'] = $mongoHost;
        }
        return $this->mongo->command($command);
    }

    /**
     * Execute command from given query
     * @param $command
     */
    public function command($command){
        return $this->mongo->command($command);
    }

    public function getIndexInfo($collection){
        return $this->mongo->getCollection($collection)->listIndexes();
    }

    public function createIndex($collection, $key, $options){
        return $this->mongo->getCollection($collection)->createIndex($key, $options);
    }

    public function findByPagination(FilterCriteria $filterCriteria, $order = null, $columns = null) {
        $this->loadCurrentDataSource($filterCriteria);
        $rows = $this->retrieveRowsPagination($filterCriteria,$order,$columns);
        return new DataSet($filterCriteria->getDataSourceName(), $rows);
    }

    protected function retrieveRowsPagination(FilterCriteria $filterCriteria, $order = null , $columns = null) {
        $criteria = $this->processNormalFilterConditions($filterCriteria,$columns);
        $sort = $this->processSortPagination($order);
        $limit = $filterCriteria->getPageSize();
        $page = $filterCriteria->getPage();
        $numSkip = $limit * ($page - 1);
        $rows = $this->mongo->findPagination($filterCriteria->getDataSourceName(), $criteria, $sort , $limit, $numSkip);
        return $this->prepareRowsForResultSet($rows);
    }

    protected function processSortPagination($order = null) {
        $sort = $this->queryBuilder->getSort($order);
        return $sort ?: array();
    }


    /**
     * Return a populated dataSet.
     *
     * Verify if the dataSet contains a ALL_DATA's flag and populate it.
     *
     * @param DataSet $dataSet
     *
     * @return DataSet $dataSet
     */
    public function populateDataSet(DataSet $dataSet)
    {

    }

    /**
     * Return a filterCriteria.
     *
     * Build a filterCriteria based on dataSourceFilter.
     *
     * @param $rows
     * @param $value
     * @param $column
     *
     * @return FilterCriteria $filterCriteria
     */
    public function buildAllDataFilter($rows, $value, $column, $dataSourceName)
    {

    }

    public function findNext($collectionName,$params,$decodeBase64 = true){
        $mongoQuery = $this->createMongoQuery($params, $decodeBase64);
        $mongoOptions = $this->createMongoOptions($params);
        return $this->findMany($collectionName, $mongoQuery, $mongoOptions);
    }

    public function findMany($collectionName, $query, $queryOptions) {
        $collection = $this->mongo->getCollection($collectionName);
        $dataFound = $collection->find($query, $queryOptions);
        $rows = $dataFound->toArray();
        return $this->prepareRowsForResultSet($rows);
    }

    public function countNext($collectionName,$params,$decodeBase64 = true){
        $mongoQuery = $this->createMongoQuery($params, $decodeBase64);
        $mongoOptions = [];
        return $this->countMany($collectionName, $mongoQuery, $mongoOptions);
    }

    public function countMany($collectionName, $query, $queryOptions) {
        $collection = $this->mongo->getCollection($collectionName);
        return $collection->count($query, $queryOptions);
    }
    /**
     * @return string|number
     */
    private function getPage($params)
    {
        return !empty($params['page']) ? $params['page'] : 1;
    }

    /**
     * @return string|number
     */
    private function getLimit($params)
    {
        return !empty($params['limit']) ? $params['limit'] : 999999;
    }

    /**
     * @return string|array
     */
    private function getOrder($params)
    {
        return !empty($params['order']) ? $params['order'] : array();
    }

    /**
     * @return array|null
     */
    private function getSearch($params)
    {
        return !empty($params['search']) ? $params['search'] : null;
    }

    /**
     * @return array
     */
    private function getSearchIn($params)
    {
        return !empty($params['search_in']) ? $params['search_in'] : array();
    }

    /**
     * @return array|null
     */
    private function getFind()
    {
        return !empty($_REQUEST['find']) ? $_REQUEST['find'] : null;
    }

    /**
     * @return string
     */
    private function getDynamicFilter($params)
    {
        return !empty($params['dynamic_filter']) ? $params['dynamic_filter'] : '';
    }

    protected function getRelation($filter, $i, $lenFilter) {
        $relation  = $filter[$i]['relation'];
        $relationOpMongo = '';
        if($i == 0 && $lenFilter > 1) {
            $relationOpMongo = $filter[$i + 1]['relation'] == 'OR' ? '$or' : '$and';
        } else if ($relation == 'AND' || $lenFilter == 1) {
            $relationOpMongo = '$and';
        } else if($relation == 'OR') {
            $relationOpMongo = '$or';
        } else {
            throw new Exception('relation key is not and neither or');
        }
        return $relationOpMongo;
    }

    /**
     * @param $dynamicFilter
     * @return array
     */
    public function createQueryFromDynamicFilter ($dynamicFilter): array
    {
        $finalQuery = array();
        $lenDynamicFilter = count($dynamicFilter);
        if ($lenDynamicFilter > 1) {
            $finalQuery['$and'] = array();
        }
        foreach ($dynamicFilter as $column => $filter) {
            $lenFilter = count($filter);
            $query = array();
            for ($i = 0; $i < $lenFilter; $i++) {
                if(!empty($filter[$i]['value'])) {
                    $value     = $filter[$i]['value'];
                    $operation = $filter[$i]['operation'];
                    $relationOpMongo = $this->getRelation($filter, $i, $lenFilter);
                    if (!array_key_exists($relationOpMongo, $query)) {
                        $query[$relationOpMongo] = array();
                    }
                    if (in_array($operation, array('IN', 'NOT_IN'))) {
                        if(!is_array($value)) {
                            $pieces = explode(";", $value);
                        } else {
                            $pieces = $value;
                        }
                        $query[$relationOpMongo][] = $operation == 'IN' ? array($column => array('$in' => $pieces)) : array($column => array('$not' => array('$in' => $pieces)));
                    } else if ($operation == 'BETWEEN') {
                        $betValue = is_string($value) ? explode(';', $value) : $value;
                        $query[$relationOpMongo][] = array($column =>array('$gte' => $betValue[0], '$lte' => $betValue[1]));
                    } else if (in_array($operation, array('CONTAINS', 'NOT_CONTAINS'))) {
                        $regex = new Regex($value);
                        $query[$relationOpMongo][] = $operation == 'CONTAINS' ? array($column => array('$regex' => $regex)) : array($column => array('$not' => $regex));
                    } else {
                        $mongoOperator = $this->opreationSimple($operation);
                        $query[$relationOpMongo][] = array($column => array($mongoOperator => $value));
                    }
                }
            }
            if ($lenDynamicFilter > 1) {
                $finalQuery['$and'][] = $query;
            } else {
                $finalQuery = $query;
            }
        }
        return $finalQuery;
    }


    /**
     * @param search
     * @param searchIn
     * @return array
     */
    protected function createSearchInQuery($search, $searchIn, $queryFilters) {
        if (!empty($searchIn)) {
            if (!array_key_exists('$or', $queryFilters)) {
                $queryFilters['$or'] = array();
            }
            foreach ($searchIn as $cols) {
                $regex = new Regex($search);
                $queryFilters['$or'][] = array($cols => array('$regex' => $regex));
            }
        }
        return $queryFilters;
    }

    /**
     * @param array $params
     * @return array
     */

    public function createMongoQuery($params, $decodeBase64 = true) {
        $search = $this->getSearch($params);
        $searchIn = $this->getSearchIn($params);
        $dynamicFilter = $this->getDynamicFilter($params);
        if ($decodeBase64) {
            $dynamicFilter = utf8_encode(base64_decode($dynamicFilter));
            $decodedDynamicFilter = empty($dynamicFilter) ? array() : json_decode($dynamicFilter, true);
        } else {
            $decodedDynamicFilter = empty($dynamicFilter) ? array() : $dynamicFilter;
        }
        $queryFilters = empty($decodedDynamicFilter) ? array() : $this->createQueryFromDynamicFilter($decodedDynamicFilter);
        $queryFilters = $this->createSearchInQuery($search, $searchIn, $queryFilters);
        return $queryFilters;
    }


    /**
     * @param array params
     * @return array
     */
    public function createMongoOptions($params) {
        $page = $this->getPage($params);
        $limit = $this->getLimit($params);
        $orderBy = $this->getOrder($params);
        $limit = (int)$limit;
        $skip = ($page - 1) * $limit;
        $order = array();
        for ($i = 0; $i < count($orderBy); $i++) {
            $content = $orderBy[$i];
            $pieces = explode('.', $content);
            $order[$pieces[0]] = $pieces[1] == 'asc' ? 1 : -1;
        }
        $optionsQuery = array('sort' => $order, 'limit' => $limit, 'skip' => $skip);
        return $optionsQuery;
    }

    /**
     * @param $operation
     * @return string
     */
    private function opreationSimple($operation): string
    {
        if ($operation == 'EQUALS') {
            $mongoOperator = '$eq';
        } else if ($operation == 'NOT_EQUALS') {
            $mongoOperator = '$ne';
        } else if ($operation == 'LESS_THAN') {
            $mongoOperator = '$lt';
        } else if ($operation == 'LESS_THAN_EQUALS') {
            $mongoOperator = '$lte';
        } else if ($operation == 'GREATER_THAN') {
            $mongoOperator = '$gt';
        } else if ($operation == 'GREATER_THAN_EQUALS') {
            $mongoOperator = '$gte';
        }
        return $mongoOperator;
    }
}
